# -*- coding: utf-8 -*-
from copy import deepcopy
from typing import Optional, List, Dict

from pip_services3_commons.commands import ICommandable, CommandSet
from pip_services3_commons.config import IConfigurable, ConfigParams
from pip_services3_commons.data import FilterParams, PagingParams
from pip_services3_commons.refer import IReferenceable, IReferences, Descriptor

from .FireMapCommandSet import FireMapCommandSet
from .IFireMapController import IFireMapController
from ..data.version1 import LocationV1, FireMapTileV1, FireMapUpdateV1, FireMapTileStatusV1, FireMapZoomV1, \
    FireMapUpdateTypeV1
from ..data.version1.FireMapTileBlockV1 import FireMapTileBlockV1
from ..map_utils import FireMapUtil
from ..persistence.IFireMapPersistence import IFireMapPersistence


class FireMapController(IFireMapController, IConfigurable, IReferenceable, ICommandable):

    def __init__(self):
        self.__persistence: IFireMapPersistence = None
        self.__command_set: FireMapCommandSet = None

    def configure(self, config: ConfigParams):
        pass

    def set_references(self, references: IReferences):
        self.__persistence = references.get_one_required(Descriptor(
            'eic-stopfires-services-firemap', 'persistence', '*', '*', '1.0'))

    def get_command_set(self) -> CommandSet:
        self.__command_set = self.__command_set or FireMapCommandSet(self)
        return self.__command_set

    def get_tiles(self, correlation_id: Optional[str], of: Optional[LocationV1], to: Optional[LocationV1],
                  zoom: Optional[int]) -> List[FireMapTileV1]:
        filter_params = FilterParams.from_tuples(
            'flng', None if not of else of.long,
            'flat', None if not of else of.lat,
            'tlng', None if not to else to.long,
            'tlat', None if not to else to.lat,
            'zoom', zoom
        )

        page = self.__persistence.get_tile_blocks_by_filter(correlation_id, filter_params, PagingParams())
        tiles = self.__tiles_blocks_to_tiles(page.data)

        return tiles

    def update_tiles(self, correlation_id: Optional[str], updates: List[FireMapUpdateV1]):
        coordinates = []
        updated_tiles: List[FireMapTileV1] = []

        for update in updates:
            coordinates.append(f'{update.lat};{update.long}')

        # get tile blocks for update
        update_blocks = self.__persistence.get_tile_blocks_by_filter(correlation_id,
                                                                     FilterParams.from_tuples('coordinates',
                                                                                              coordinates),
                                                                     PagingParams()).data

        absent_zooms: List[int] = [e.value for e in FireMapZoomV1]

        for update in updates:
            current_update_blocks = []

            for block in update_blocks:
                # check if update coordinates in block coordinates
                if block.flng <= update.long <= block.tlng and block.tlat <= update.lat <= block.flat:
                    # find absent zooms
                    if FireMapZoomV1.Zoom100km.value != block.zoom:
                        current_update_blocks.append(block)
                    absent_zooms = list(filter(lambda z: z != block.zoom, absent_zooms))

            # if block is not created, or some absent
            if len(absent_zooms) > 0:
                new_blocks = self.__create_blocks(update.lat, update.long, absent_zooms)
                current_update_blocks.extend(new_blocks)

            # update blocks for all zooms
            for block in current_update_blocks:
                # update current based on states of the children
                if update.type is not None and not self.__can_update_block(block, update_blocks, update.type):
                    continue

                # compose id by zoom
                loc = FireMapUtil.get_locator(LocationV1(update.lat, update.long))
                id = FireMapUtil.compose_tile_loc_by_zoom(loc, block.zoom)

                if update.type == FireMapUpdateTypeV1.CLEAR:
                    block.clear[id].updated = update.time
                    block.clear[id].drone_id = update.drone_id
                    block.clear[id].people = update.people
                elif update.type == FireMapUpdateTypeV1.SMOKE:
                    block.smoke[id].updated = update.time
                    block.smoke[id].drone_id = update.drone_id
                    block.smoke[id].people = update.people
                elif update.type == FireMapUpdateTypeV1.FIRE:
                    block.fire[id].updated = update.time
                    block.fire[id].drone_id = update.drone_id
                    block.fire[id].people = update.people

                # update peoples count for all statuses
                block.clear[id].people_updated = update.time
                block.clear[id].people = update.people
                block.smoke[id].people_updated = update.time
                block.smoke[id].people = update.people
                block.fire[id].people_updated = update.time
                block.fire[id].people = update.people

            update_blocks.extend(current_update_blocks)

            # TODO: add update listener

        self.__persistence.update_tile_blocks(correlation_id, update_blocks)

    def __create_blocks(self, lat: float, long: float, zooms: List[int]) -> List[FireMapTileBlockV1]:
        blocks: List[FireMapTileBlockV1] = []
        zooms = zooms or [e.value for e in FireMapZoomV1]

        locator = FireMapUtil.get_locator(LocationV1(lat, long))

        for tile_zoom in zooms:
            block_zoom = tile_zoom + 1
            block_coord = FireMapUtil.get_locator_coordinates(locator, block_zoom)
            # calc id as block locator
            center = FireMapUtil.calc_center_coordinates(LocationV1(block_coord[0].lat, block_coord[0].long),
                                                         LocationV1(block_coord[1].lat, block_coord[1].long))

            loc_of_center = FireMapUtil.get_locator(LocationV1(center.lat, center.long))
            loc_of_block = (FireMapUtil.compose_tile_loc_by_zoom(loc_of_center, block_zoom) + '_' + str(
                block_zoom - 1)).upper()

            block = FireMapTileBlockV1(
                id=loc_of_block,
                zoom=tile_zoom,
                flat=block_coord[0].lat,
                flng=block_coord[0].long,
                tlat=block_coord[1].lat,
                tlng=block_coord[1].long,
                clear={},
                smoke={},
                fire={},
            )

            # if it biggest block
            iterations = 0

            if tile_zoom == FireMapZoomV1.Zoom50m:
                iterations = 24  # 576 tiles for 'AA-XX'
            elif tile_zoom == FireMapZoomV1.Zoom1km:
                iterations = 10  # 100 tiles for '00-99'
            elif tile_zoom == FireMapZoomV1.Zoom20km:
                iterations = 24  # 576 tiles for 'AA-XX'
            elif tile_zoom == FireMapZoomV1.Zoom50km:
                iterations = 10  # 100 tiles for '00-99'
            elif tile_zoom == FireMapZoomV1.Zoom100km:
                iterations = 18  # 324 tiles for 'AA-XX'

            # fill all current block tiles
            for i in range(iterations):
                for j in range(iterations):
                    tile_locator = FireMapUtil.get_loc_by_zoom(locator, block_zoom)

                    if len(tile_locator) in [2, 6]:
                        tile_locator += str(i) + str(j)
                    elif len(tile_locator) in [0, 4, 8]:
                        tile_locator += chr(i + 65) + chr(j + 65)

                    tile_coord = FireMapUtil.get_locator_coordinates(tile_locator, tile_zoom)

                    # calc tile id as locator
                    id = FireMapUtil.compose_tile_loc_by_zoom(tile_locator, tile_zoom)

                    tile_status = FireMapTileStatusV1(
                        id=id,
                        flat=tile_coord[0].lat,
                        flng=tile_coord[0].long,
                        tlat=tile_coord[1].lat,
                        tlng=tile_coord[1].long,
                        updated=None,
                        drone_id=None,
                        people=None,
                        people_updated=None
                    )

                    block.clear[id] = deepcopy(tile_status)
                    block.smoke[id] = deepcopy(tile_status)
                    block.fire[id] = deepcopy(tile_status)

            blocks.append(block)

        return blocks

    def __tiles_blocks_to_tiles(self, tiles_blocks: List[FireMapTileBlockV1]) -> List[FireMapTileV1]:
        tiles: Dict[str, FireMapTileV1] = {}

        for block in tiles_blocks:
            for id in block.clear.keys():
                if tiles.get(id) is None:
                    tiles[id] = FireMapTileV1(
                        id=id,
                        zoom=block.zoom,
                        flng=block.clear[id].flng,
                        flat=block.clear[id].flat,
                        tlng=block.clear[id].tlng,
                        tlat=block.clear[id].tlat,

                        clear=block.clear[id].updated,
                        cdrone=block.clear[id].drone_id,
                        smoke=block.smoke[id].updated,
                        sdrone=block.smoke[id].drone_id,
                        fire=block.fire[id].updated,
                        fdrone=block.fire[id].drone_id,

                        people=block.fire[id].people_updated,
                        pdrone=block.fire[id].drone_id,
                        people_count=block.fire[id].people
                    )
                else:
                    tiles[id].clear = block.clear[id].updated
                    tiles[id].cdrone = block.clear[id].drone_id
                    tiles[id].smoke = block.smoke[id].updated
                    tiles[id].sdrone = block.smoke[id].drone_id
                    tiles[id].fire = block.fire[id].updated
                    tiles[id].fdrone = block.fire[id].drone_id

                    tiles[id].people = block.fire[id].people_updated
                    tiles[id].pdrone = block.fire[id].drone_id
                    tiles[id].people_count = block.fire[id].people

        return list(tiles.values())

    def __can_update_block(self, curr_block: FireMapTileBlockV1, blocks: List[FireMapTileBlockV1],
                           type: str) -> bool:
        child_blocks = filter(lambda b: b.zoom < curr_block.zoom, blocks)
        checked = 0
        count_of_tiles = 0

        for child_block in child_blocks:
            count_of_tiles += len(child_block.clear)

            for id in child_block.clear:
                # if settled values updated
                if child_block.clear[id].updated and child_block.smoke[id].updated and child_block.clear[id].updated:
                    if type == FireMapUpdateTypeV1.CLEAR:
                        if (child_block.clear[id].updated >= child_block.smoke[id].updated and
                                child_block.clear[id].updated >= child_block.fire[id].updated):
                            checked += 1
                    elif type == FireMapUpdateTypeV1.SMOKE:
                        if child_block.smoke[id].updated >= child_block.fire[id].updated:
                            checked += 1
                    elif type == FireMapUpdateTypeV1.FIRE:
                        return True

        # if checked all blocks
        return count_of_tiles == checked
